# TODO: Adjustable line length? (BSB wraps lines at 40-43 characters)
def list_multiline_verse(verse):
    lines = []
    
    next_line = verse
    
    while len(next_line) > 80:
        # Split the verse if there's more than one line left
        space_split = next_line[:79].rfind(' ')
        lines.append(next_line[:space_split])
        next_line = next_line[space_split:].lstrip()
            
    # Append last line of verse
    lines.append(next_line)

    return lines


# TODO: Replace consecutive spaces with single spaces
# TODO: Input line length?
def verses_to_wall_of_text(verse_records, verse_numbers=False): 
    verses = ''
    for row in verse_records:
        # Skip empty verses so orphaned verse numbers or extra whitespace
        # is not displayed
        if not row['text']:
            continue
        if verse_numbers:
            verses += str(row['verse']) + ' '
        
        verses += row['text'].strip() + ' '
    
    verses_split = list_multiline_verse(verses)
    wrapped_verses = '\n'.join(verses_split)
    
    return wrapped_verses


def create_link_label(translation, book, chapter=None, verse=None):
    """Creates a link label, eg. `Isaiah 14:12-20`
    """
    label = book
    
    if chapter:
        label += f" {chapter}"
        
        if verse:
            label += f":{verse}"
    
    label += f" {translation}"
    
    return label


# TODO: Print paragraphs from Bible format
def create_markdown_excerpt(bible_client, verse_records, book, chapter, verse):
    """Generate Markdown excerpt for the verses.

    Args:
        verse_records (_type_): _description_
        params (_type_): _description_
    """
    verse_text = verses_to_wall_of_text(verse_records)
    book = bible_client.get_book_from_abbreviation(book)
    output = (
        '###\n'
        '\n______________________________________________________________________\n'
        f"\n{verse_text}"
        f"\n([{create_link_label(bible_client.translation, book, chapter, verse)}]"
        f"({bible_client.create_link(book, chapter, verse)}))\n"
        '\n______________________________________________________________________'
    )
    return output
    

# TODO: Return paragraphs from Bible format
def render_reference_results(bible_client, format, verse_records, verse_numbers=False, book=None, chapter=None, verse=None):
    """_summary_

    Args:
        bible_client (BibleClient): Used to create the link and label if needed.
        format (_type_): _description_
        verse_records (_type_): _description_
        verse_numbers (bool, optional): _description_. Defaults to False.
        book (_type_, optional): _description_. Defaults to None.
        chapter (_type_, optional): _description_. Defaults to None.
        verse (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    match format: 
        case 'txt':
            return verses_to_wall_of_text(verse_records, verse_numbers)

        case 'md':
            return create_markdown_excerpt(bible_client, verse_records, book, chapter, verse)


def render_search_results(
    bible_client,
    verse_records,
    phrase,
    testament=None,
    book=None,
    chapter=None
):
    if chapter:
        book = bible_client.get_book_from_abbreviation(book)
        output = f"{len(verse_records)} occurrences of '{phrase}' in {book} {chapter} ({bible_client.translation}):\n___\n"
        for verse in verse_records:
            output += f"\n{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n"
        
    elif book:
        book = bible_client.get_book_from_abbreviation(book)
        output = f"{len(verse_records)} occurrences of '{phrase}' in {book} ({bible_client.translation}):\n___\n"
        for verse in verse_records:
            output += f"\n{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n"
            
    elif testament:
        testament = 'New Testament' if testament == 'nt' else 'Old Testament'
        output = f"{len(verse_records)} occurrences of '{phrase}' in the {testament} ({bible_client.translation}):\n___\n"
        for verse in verse_records:
            output += f"\n{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n"
            
    else:
        output = f"{len(verse_records)} occurrences of '{phrase}' in the {bible_client.translation} Bible:\n___\n"
        for verse in verse_records:
            output += f"\n{verse['book']} {verse['chapter']}:{verse['verse']}:\n{verse['text']}\n___\n"

    return output



import urllib.request

import pytest

from berea.bible import BibleClient, BibleInputError



def valid_url(url):
    try:
        with urllib.request.urlopen(url) as response:
            if response.status == 200:
                return True
            else:
                return False
    except urllib.error.HTTPError as e:
        print("HTTP error:", e.code)
        return False
    except urllib.error.URLError as e:
        print("URL error:", e.reason)
        return False


@pytest.mark.parametrize(
    "msg, book, chapter, verse, translation, expected_link",
    [
        (
            "Creating link for a single verse failed",
            "John", "3", "16", "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=john.3.16&options=NVHUG"
        ),
        (
            "Creating link for multiple verses failed",
            "John", "3", "16-18", "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=john.3.16-john.3.18&options=NVHUG"
        ),
        (
            "Creating link for a chapter failed",
            "Psalms", "117", None, "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=psalm.117&options=NVHUG"
        ),
        (
            "Creating link for a book",
            "III John", None, None, "BSB",
            "https://www.stepbible.org/?q=version=BSB@reference=3john&options=NVHUG"
        ),
    ]
)
def test_create_link(msg, book, chapter, verse, translation, expected_link):
    bible = BibleClient(translation)
    actual_link = bible.create_link(book, chapter, verse)
    assert actual_link == expected_link, msg
    assert valid_url(actual_link)


def test_validate_resource_abbreviations():
    """Creating links from the resource's book abbreviations yields valid URLs."""
    bible = BibleClient('BSB')
    
    books = bible.get_bible_cursor().execute("SELECT * FROM books").fetchall()
    
    for book in books:
        link = bible.create_link(book['name'])
        assert valid_url(link), f"{book['name']} produced invalid link: {link}"


@pytest.mark.parametrize(
    "translation",
    [
        ('KJV'),
        ('BSB'),
        ('RLT'),        
        ('UKJV'),        
    ]
)
def test_create_bible_db(translation):
    bible = BibleClient(translation)
    bible.create_bible_db()

    cursor = bible.get_bible_cursor()

    msg = 'Downloading the translation database failed.'
    assert pytest.translation_exists(translation), msg

    sql = "SELECT name FROM sqlite_master WHERE type='table';"
    actual_tables = [row['name'] for row in cursor.execute(sql).fetchall()]

    table_record_counts = {
        'books': 66,
        'verses': 31102,
    }

    msg = 'Renaming the database tables failed.'
    renamed_tables = table_record_counts.keys()
    assert set(renamed_tables).issubset(actual_tables), msg

    for table, expected_records_count in table_record_counts.items():
        sql = f"SELECT COUNT(*) FROM {table};"
        actual_records_count = cursor.execute(sql).fetchone()[0]
        
        msg = f"'{table}' table does not contain expected record count."
        assert actual_records_count == expected_records_count, msg

    created_tables = ['abbreviations', 'resources', 'resources_abbreviations']
    for expected_table in created_tables:
        msg =  f"'{expected_table}' table does not exist."
        assert expected_table in actual_tables, msg


def test_search_testament_error():
    """This error is only reachable from BibleClient.
    """
    bible = BibleClient('BSB')
    
    with pytest.raises(BibleInputError, match="Invalid testament='secret'."):
        bible.search_testament("That's all folks!", "secret")



import pytest

from berea.render import list_multiline_verse


@pytest.mark.parametrize(
    "verse, verse_list",
    [   
        # John 3:16 KJV (141 characters)
        (
            "For God so loved the world, that he gave his only begotten Son, that whosoever believeth in him should not perish, but have everlasting life.",
            [
                "For God so loved the world, that he gave his only begotten Son, that whosoever",
                "believeth in him should not perish, but have everlasting life."
            ]
        ),
        # Esther 8:9 KJV (528 characters)
        (
            "Then were the king’s scribes called at that time in the third month, that is, the month Sivan, on the three and twentieth day thereof; and it was written according to all that Mordecai commanded unto the Jews, and to the lieutenants, and the deputies and rulers of the provinces which are from India unto Ethiopia, an hundred twenty and seven provinces, unto every province according to the writing thereof, and unto every people after their language, and to the Jews according to their writing, and according to their language.",
            [
                "Then were the king’s scribes called at that time in the third month, that is,",
                "the month Sivan, on the three and twentieth day thereof; and it was written",
                "according to all that Mordecai commanded unto the Jews, and to the",
                "lieutenants, and the deputies and rulers of the provinces which are from India",
                "unto Ethiopia, an hundred twenty and seven provinces, unto every province",
                "according to the writing thereof, and unto every people after their language,",
                "and to the Jews according to their writing, and according to their language."
            ]
        ),
    ]
)
def test_list_multiline_verse(verse, verse_list):
    assert list_multiline_verse(verse) == verse_list
